@@ -12,7 +12,7 @@ from typing import (
     TYPE_CHECKING,
 )
 
-from .exceptions import NoSuchUserError
+from .exceptions import InvalidSessionError, NoSuchUserError
 from .protocol.messages import (
     AddHatedInterest,
     AddInterest,
@@ -79,6 +79,7 @@ RT = TypeVar('RT')
 
 Recommendations = Tuple[List[Recommendation], List[Recommendation]]
 UserInterests = Tuple[List[str], List[str]]
+SharesReply = Tuple[List[DirectoryData], List[DirectoryData]]
 
 
 class UserStatusInfo(NamedTuple):
@@ -258,7 +259,8 @@ class GrantRoomMembershipCommand(BaseCommand[PrivateRoomGrantMembership.Response
             }
         )
 
-    def handle_response(self, client: SoulSeekClient, response: PrivateRoomGrantMembership.Response) -> Tuple[Room, User]:
+    def handle_response(
+            self, client: SoulSeekClient, response: PrivateRoomGrantMembership.Response) -> Tuple[Room, User]:
         return (
             client.rooms.get_or_create_room(response.room),
             client.users.get_user_object(response.username)
@@ -286,7 +288,8 @@ class RevokeRoomMembershipCommand(BaseCommand[PrivateRoomRevokeMembership.Respon
             }
         )
 
-    def handle_response(self, client: SoulSeekClient, response: PrivateRoomRevokeMembership.Response) -> Tuple[Room, User]:
+    def handle_response(
+            self, client: SoulSeekClient, response: PrivateRoomRevokeMembership.Response) -> Tuple[Room, User]:
         return (
             client.rooms.get_or_create_room(response.room),
             client.users.get_user_object(response.username)
@@ -308,10 +311,7 @@ class DropRoomMembershipCommand(BaseCommand[PrivateRoomMembershipRevoked.Respons
         return ExpectedResponse(
             ServerConnection,
             PrivateRoomMembershipRevoked.Response,
-            fields={
-                'room': self.room,
-                'username': client.session.user.name
-            }
+            fields={'room': self.room}
         )
 
     def handle_response(self, client: SoulSeekClient, response: PrivateRoomMembershipRevoked.Response) -> Room:
@@ -348,7 +348,8 @@ class GetItemRecommendationsCommand(BaseCommand[GetItemRecommendations.Response,
             }
         )
 
-    def handle_response(self, client: SoulSeekClient, response: GetItemRecommendations.Response) -> List[Recommendation]:
+    def handle_response(
+            self, client: SoulSeekClient, response: GetItemRecommendations.Response) -> List[Recommendation]:
         return response.recommendations
 
 
@@ -409,7 +410,7 @@ class GetItemSimilarUsersCommand(BaseCommand[GetItemSimilarUsers.Response, List[
         return list(map(client.users.get_user_object, response.usernames))
 
 
-class GetSimilarUsersCommand(BaseCommand[GetSimilarUsers.Response, List[User]]):
+class GetSimilarUsersCommand(BaseCommand[GetSimilarUsers.Response, List[Tuple[User, int]]]):
 
     async def send(self, client: SoulSeekClient):
         await client.network.send_server_messages(
@@ -422,14 +423,20 @@ class GetSimilarUsersCommand(BaseCommand[GetSimilarUsers.Response, List[User]]):
             GetSimilarUsers.Response
         )
 
-    def handle_response(self, client: SoulSeekClient, response: GetSimilarUsers.Response) -> List[User]:
-        return [
-            client.users.get_user_object(user.username)
-            for user in response.users
-        ]
+    def handle_response(self, client: SoulSeekClient, response: GetSimilarUsers.Response) -> List[Tuple[User, int]]:
+        similar_users = []
+        for similar_user in response.users:
+            similar_users.append(
+                (
+                    client.users.get_user_object(similar_user.username),
+                    similar_user.score
+                )
+            )
+
+        return similar_users
 
 
-class GetPeerAddressCommand(BaseCommand[GetPeerAddress.Response, Tuple[str, int, int]]):
+class GetPeerAddressCommand(BaseCommand[GetPeerAddress.Response, Tuple[str, int, Optional[int]]]):
 
     def __init__(self, username: str):
         self.username: str = username
@@ -448,7 +455,8 @@ class GetPeerAddressCommand(BaseCommand[GetPeerAddress.Response, Tuple[str, int,
             }
         )
 
-    def handle_response(self, client: SoulSeekClient, response: GetPeerAddress.Response) -> Tuple[str, int, int]:
+    def handle_response(
+            self, client: SoulSeekClient, response: GetPeerAddress.Response) -> Tuple[str, int, Optional[int]]:
         return (response.ip, response.port, response.obfuscated_port)
 
 
@@ -509,7 +517,7 @@ class TogglePublicChatCommand(BaseCommand[None, None]):
 
     async def send(self, client: SoulSeekClient):
         if self.enable:
-            message = EnablePublicChat.Request()
+            message: MessageDataclass = EnablePublicChat.Request()
         else:
             message = DisablePublicChat.Request()
         await client.network.send_server_messages(message)
@@ -653,6 +661,9 @@ class RoomMessageCommand(BaseCommand[RoomChatMessage.Response, RoomMessage]):
         )
 
     def build_expected_response(self, client: SoulSeekClient) -> Optional[ExpectedResponse]:
+        if not client.session:  # pragma: no cover
+            raise InvalidSessionError("user is not logged in")
+
         return ExpectedResponse(
             ServerConnection,
             RoomChatMessage.Response,
@@ -664,6 +675,9 @@ class RoomMessageCommand(BaseCommand[RoomChatMessage.Response, RoomMessage]):
         )
 
     def handle_response(self, client: SoulSeekClient, response: RoomChatMessage.Response) -> RoomMessage:
+        if not client.session:  # pragma: no cover
+            raise InvalidSessionError("user is not logged in")
+
         return RoomMessage(
             timestamp=int(time.time()),
             user=client.users.get_user_object(client.session.user.name),
@@ -687,6 +701,9 @@ class SetRoomTickerCommand(BaseCommand[RoomTickerAdded.Response, None]):
         )
 
     def build_expected_response(self, client: SoulSeekClient) -> Optional[ExpectedResponse]:
+        if not client.session:  # pragma: no cover
+            raise InvalidSessionError("user is not logged in")
+
         return ExpectedResponse(
             ServerConnection,
             RoomTickerAdded.Response,
@@ -883,7 +900,7 @@ class PeerGetUserInfoCommand(BaseCommand[PeerUserInfoReply.Request, UserInfo]):
         )
 
 
-class PeerGetSharesCommand(BaseCommand[PeerSharesReply.Request, Tuple[List[DirectoryData], List[DirectoryData]]]):
+class PeerGetSharesCommand(BaseCommand[PeerSharesReply.Request, SharesReply]):
 
     def __init__(self, username: str):
         self.username: str = username
@@ -901,7 +918,7 @@ class PeerGetSharesCommand(BaseCommand[PeerSharesReply.Request, Tuple[List[Direc
         )
 
     def handle_response(
-            self, client: SoulSeekClient, response: PeerSharesReply.Request) -> Tuple[List[DirectoryData], List[DirectoryData]]:
+            self, client: SoulSeekClient, response: PeerSharesReply.Request) -> SharesReply:
         locked_dirs = response.locked_directories or []
         return response.directories, locked_dirs
 
@@ -930,5 +947,6 @@ class PeerGetDirectoryContentCommand(BaseCommand[PeerDirectoryContentsReply.Requ
             }
         )
 
-    def handle_response(self, client: SoulSeekClient, response: PeerDirectoryContentsReply.Request) -> List[DirectoryData]:
+    def handle_response(
+            self, client: SoulSeekClient, response: PeerDirectoryContentsReply.Request) -> List[DirectoryData]:
         return response.directories
