@@ -6,7 +6,11 @@ import logging
 from typing import Deque, List, Optional, Union, Tuple
 
 from .base_manager import BaseManager
-from .constants import POTENTIAL_PARENTS_CACHE_SIZE
+from .constants import (
+    DEFAULT_PARENT_MIN_SPEED,
+    DEFAULT_PARENT_SPEED_RATIO,
+    POTENTIAL_PARENTS_CACHE_SIZE,
+)
 from .events import (
     on_message,
     build_message_map,
@@ -27,12 +31,14 @@ from .protocol.messages import (
     DistributedChildDepth,
     DistributedSearchRequest,
     DistributedServerSearchRequest,
+    GetUserStats,
     MessageDataclass,
     MinParentsInCache,
     ParentInactivityTimeout,
     ParentMinSpeed,
     ParentSpeedRatio,
     PotentialParents,
+    ResetDistributed,
     ServerSearchRequest,
     ToggleParentSearch,
 )
@@ -54,8 +60,12 @@ logger = logging.getLogger(__name__)
 
 @dataclass
 class DistributedPeer:
+    """Represents a distributed peer and its values in the distributed network"""
     username: str
     connection: Optional[PeerConnection] = None
+    """Distributed connection (type=D) associated with the peer. A `None` value
+    indicates the peer represents the current client
+    """
     branch_level: Optional[int] = None
     branch_root: Optional[str] = None
     child_depth: Optional[int] = None
@@ -87,6 +97,8 @@ class DistributedNetwork(BaseManager):
         self.min_parents_in_cache: Optional[int] = None
         self.parent_inactivity_timeout: Optional[int] = None
         self.distributed_alive_interval: Optional[int] = None
+        self._max_children: int = 5
+        self._accept_children: bool = True
 
         self._MESSAGE_MAP = build_message_map(self)
 
@@ -123,25 +135,43 @@ class DistributedNetwork(BaseManager):
             * level = level parent advertised + 1
             * root = whatever our parent sent us initially
         """
-        username = self._session.user.name
+        username = self._session.user.name  # type: ignore[union-attr]
         if self.parent:
             # We are the branch root
             if self.parent.branch_root == username:
                 return username, 0
             else:
-                return self.parent.branch_root, self.parent.branch_level + 1
+                return self.parent.branch_root, self.parent.branch_level + 1  # type: ignore
 
         return username, 0
 
-    def get_distributed_peer(self, username: str, connection: PeerConnection) -> DistributedPeer:
+    def get_distributed_peer(self, connection: PeerConnection) -> Optional[DistributedPeer]:
+        """Get the distributed peer object related to the given connection.
+
+        This method will return `None` if the connection was not properly
+        initialized (no username set on the connection) or if there is no peer
+        object associated with the connection
+        """
+        if connection.username is None:
+            return None
+
         for peer in self.distributed_peers:
-            if peer.username == username and peer.connection == connection:
+            if peer.username == connection.username and peer.connection == connection:
                 return peer
 
+        return None
+
     async def reset(self):
         """Disconnects all parent and child connections"""
-        await self._unset_children()
-        await self._unset_parent()
+        await self._disconnect_children()
+        await self._disconnect_parent()
+
+    def _reset_server_values(self):
+        self.parent_min_speed = None
+        self.parent_speed_ratio = None
+        self.distributed_alive_interval = None
+        self.min_parents_in_cache = None
+        self.parent_inactivity_timeout = None
 
     async def _set_parent(self, peer: DistributedPeer):
         """Sets the given peer to be our parent.
@@ -187,17 +217,23 @@ class DistributedNetwork(BaseManager):
                 if peer.connection:
                     await peer.connection.disconnect(reason=CloseReason.REQUESTED)
 
-    async def _unset_children(self):
-        await asyncio.gather(*[
-                self._unset_child(child) for child in self.children
-            ],
+    async def _disconnect_children(self):
+        await asyncio.gather(
+            *[self._disconnect_child(child) for child in self.children],
             return_exceptions=True
         )
 
-    async def _unset_child(self, peer: DistributedPeer):
+    async def _disconnect_child(self, peer: DistributedPeer):
         if peer.connection:
             await peer.connection.disconnect(CloseReason.REQUESTED)
 
+    async def _disconnect_parent(self):
+        if self.parent:
+            if self.parent.connection:
+                await self.parent.connection.disconnect(CloseReason.REQUESTED)
+            else:
+                await self._unset_parent()
+
     async def _unset_parent(self):
         logger.debug(f"unset parent {self.parent!r}")
 
@@ -224,27 +260,11 @@ class DistributedNetwork(BaseManager):
         root, level = self._get_advertised_branch_values()
         logger.info(f"notifying server of our parent : level={level} root={root}")
 
-        messages = [
+        await self._network.send_server_messages(*[
             BranchLevel.Request(level),
-            BranchRoot.Request(root)
-        ]
-
-        if self.parent:
-            logger.info("notifying server we are not looking for parent")
-            messages.extend([
-                ToggleParentSearch.Request(False),
-                AcceptChildren.Request(True)
-            ])
-        else:
-            logger.info("notifying server we are looking for parent")
-            # The original Windows client sends out the child depth (=0) and the
-            # ParentIP
-            messages.extend([
-                ToggleParentSearch.Request(True),
-                AcceptChildren.Request(True)
-            ])
-
-        await self._network.send_server_messages(*messages)
+            BranchRoot.Request(root),
+            ToggleParentSearch.Request(False if self.parent else True)
+        ])
 
     async def _notify_children_of_branch_values(self):
         root, level = self._get_advertised_branch_values()
@@ -259,6 +279,21 @@ class DistributedNetwork(BaseManager):
         if peer.username in self.potential_parents:
             return
 
+        if not self._accept_children:
+            logger.debug(f"not accepting children, rejecting peer as child: {peer}")
+            if peer.connection:  # Satisfy type checker
+                await peer.connection.disconnect(CloseReason.REQUESTED)
+            return
+
+        if len(self.children) >= self._max_children:
+            logger.debug(
+                f"maximum amount of children reached ({len(self.children)} / {self._max_children}), "
+                f"rejecting peer as child: {peer}"
+            )
+            if peer.connection:  # Satisfy type checker
+                await peer.connection.disconnect(CloseReason.REQUESTED)
+            return
+
         await self._add_child(peer)
 
     async def _add_child(self, peer: DistributedPeer):
@@ -296,27 +331,35 @@ class DistributedNetwork(BaseManager):
     # Server messages
 
     @on_message(ParentMinSpeed.Response)
-    async def _on_parent_min_speed(self, message: ParentMinSpeed.Response, connection: ServerConnection):
+    async def _on_parent_min_speed(
+            self, message: ParentMinSpeed.Response, connection: ServerConnection):
         self.parent_min_speed = message.speed
+        await self._request_user_stats()
 
     @on_message(ParentSpeedRatio.Response)
-    async def _on_parent_speed_ratio(self, message: ParentSpeedRatio.Response, connection: ServerConnection):
+    async def _on_parent_speed_ratio(
+            self, message: ParentSpeedRatio.Response, connection: ServerConnection):
         self.parent_speed_ratio = message.ratio
+        await self._request_user_stats()
 
     @on_message(MinParentsInCache.Response)
-    async def _on_min_parents_in_cache(self, message: MinParentsInCache.Response, connection: ServerConnection):
+    async def _on_min_parents_in_cache(
+            self, message: MinParentsInCache.Response, connection: ServerConnection):
         self.min_parents_in_cache = message.amount
 
     @on_message(DistributedAliveInterval.Response)
-    async def _on_ditributed_alive_interval(self, message: DistributedAliveInterval.Response, connection: ServerConnection):
+    async def _on_ditributed_alive_interval(
+            self, message: DistributedAliveInterval.Response, connection: ServerConnection):
         self.distributed_alive_interval = message.interval
 
     @on_message(ParentInactivityTimeout.Response)
-    async def _on_parent_inactivity_timeout(self, message: ParentInactivityTimeout.Response, connection: ServerConnection):
+    async def _on_parent_inactivity_timeout(
+            self, message: ParentInactivityTimeout.Response, connection: ServerConnection):
         self.parent_inactivity_timeout = message.timeout
 
     @on_message(PotentialParents.Response)
-    async def _on_potential_parents(self, message: PotentialParents.Response, connection: ServerConnection):
+    async def _on_potential_parents(
+            self, message: PotentialParents.Response, connection: ServerConnection):
         if not self._settings.debug.search_for_parent:
             logger.debug("ignoring PotentialParents message : searching for parent is disabled")
             return
@@ -360,22 +403,24 @@ class DistributedNetwork(BaseManager):
                 )
                 await self._set_parent(parent)
 
-        # Pass the message on to the children anyway, could just be that the
-        #
         await self.send_messages_to_children(message)
 
+    @on_message(ResetDistributed.Response)
+    async def _on_reset_distributed(self, message: ResetDistributed.Response, connection: ServerConnection):
+        await self.reset()
+
     # Distributed messages
 
     @on_message(DistributedBranchLevel.Request)
-    async def _on_distributed_branch_level(self, message: DistributedBranchLevel.Request, connection: PeerConnection):
+    async def _on_distributed_branch_level(
+            self, message: DistributedBranchLevel.Request, connection: PeerConnection):
         logger.info(f"branch level {message.level!r}: {connection!r}")
 
-        if not connection.username:
-            logger.warning(
-                "got DistributedBranchLevel for a connection that wasn't properly initialized")
+        peer = self.get_distributed_peer(connection)
+        if not peer:
+            logger.warning(f"distributed peer object not found for connection: {connection} : {message}")
             return
 
-        peer = self.get_distributed_peer(connection.username, connection)
         peer.branch_level = message.level
 
         # Branch root is not always sent in case the peer advertises branch
@@ -390,16 +435,15 @@ class DistributedNetwork(BaseManager):
             await self._notify_children_of_branch_values()
 
     @on_message(DistributedBranchRoot.Request)
-    async def _on_distributed_branch_root(self, message: DistributedBranchRoot.Request, connection: PeerConnection):
+    async def _on_distributed_branch_root(
+            self, message: DistributedBranchRoot.Request, connection: PeerConnection):
         logger.info(f"branch root {message.username!r}: {connection!r}")
 
-        if not connection.username:
-            logger.warning(
-                "got DistributedBranchRoot for a connection that wasn't properly initialized")
+        peer = self.get_distributed_peer(connection)
+        if not peer:
+            logger.warning(f"distributed peer object not found for connection: {connection} : {message}")
             return
 
-        peer = self.get_distributed_peer(connection.username, connection)
-
         # When we receive branch level 0 we automatically assume the root is the
         # peer who sent the sender
         # Don't do anything if the branch root is what we expected it to be
@@ -414,21 +458,23 @@ class DistributedNetwork(BaseManager):
             await self._notify_children_of_branch_values()
 
     @on_message(DistributedChildDepth.Request)
-    async def _on_distributed_child_depth(self, message: DistributedChildDepth.Request, connection: PeerConnection):
-        if not connection.username:
-            logger.warning(
-                "got DistributedChildDepth for a connection that wasn't properly initialized")
+    async def _on_distributed_child_depth(
+            self, message: DistributedChildDepth.Request, connection: PeerConnection):
+        peer = self.get_distributed_peer(connection)
+        if not peer:
+            logger.warning(f"distributed peer object not found for connection: {connection} : {message}")
             return
 
-        peer = self.get_distributed_peer(connection.username, connection)
         peer.child_depth = message.depth
 
     @on_message(DistributedSearchRequest.Request)
-    async def _on_distributed_search_request(self, message: DistributedSearchRequest.Request, connection: PeerConnection):
+    async def _on_distributed_search_request(
+            self, message: DistributedSearchRequest.Request, connection: PeerConnection):
         await self.send_messages_to_children(message)
 
     @on_message(DistributedServerSearchRequest.Request)
-    async def _on_distributed_server_search_request(self, message: DistributedServerSearchRequest.Request, connection: PeerConnection):
+    async def _on_distributed_server_search_request(
+            self, message: DistributedServerSearchRequest.Request, connection: PeerConnection):
         if message.distributed_code != DistributedSearchRequest.Request.MESSAGE_ID:
             logger.warning(f"no handling for server search request with code {message.distributed_code}")
             return
@@ -441,9 +487,69 @@ class DistributedNetwork(BaseManager):
         )
         await self.send_messages_to_children(dmessage)
 
+    @on_message(GetUserStats.Response)
+    async def _on_get_user_stats(self, message: GetUserStats.Response, connection: ServerConnection):
+        if self._session and message.username == self._session.user.name:
+            speed = message.user_stats.avg_speed
+
+            if self.parent_min_speed is None:
+                logger.debug(
+                    f"using default parent_min_speed: {DEFAULT_PARENT_MIN_SPEED}")
+                parent_min_speed = DEFAULT_PARENT_MIN_SPEED
+            else:
+                parent_min_speed = self.parent_min_speed
+
+            if self.parent_speed_ratio is None:
+                logger.debug(
+                    f"using default parent_speed_ratio: {DEFAULT_PARENT_SPEED_RATIO}")
+                parent_speed_ratio = DEFAULT_PARENT_SPEED_RATIO
+            else:
+                parent_speed_ratio = self.parent_speed_ratio
+
+            if speed < parent_min_speed * 1024:
+                self._accept_children = False
+                self._max_children = 0
+
+            else:
+                self._accept_children = True
+                self._max_children = self._calculate_max_children(speed, parent_speed_ratio)
+
+            logger.debug(
+                f"adjusting distributed children values: "
+                f"accept={self._accept_children}, max_children={self._max_children}"
+            )
+            await self._network.send_server_messages(
+                AcceptChildren.Request(self._accept_children)
+            )
+
+    def _has_parent_speed_values(self) -> bool:
+        """Returns `True` if `ParentMinSpeed` and `ParentSpeedRatio` has been
+        received from the server
+        """
+        return self.parent_min_speed is not None and self.parent_speed_ratio is not None
+
+    async def _request_user_stats(self):
+        """Requests the user stats for the currently logged on user a valid
+        session is available and parent speed values have been received
+        """
+        if self._session is None:
+            return
+
+        if self._has_parent_speed_values():
+            await self._network.send_server_messages(
+                GetUserStats.Request(self._session.user.name)
+            )
+
+    def _calculate_max_children(self, upload_speed: int, parent_speed_ratio: int) -> int:
+        """Calculates the maximum number of children based on the given upload
+        speed
+        """
+        divider = (parent_speed_ratio / 10) * 1024
+        return int(upload_speed / divider)
+
     async def _on_peer_connection_initialized(self, event: PeerInitializedEvent):
         if event.connection.connection_type == PeerConnectionType.DISTRIBUTED:
-            peer = DistributedPeer(event.connection.username, event.connection) # type: ignore
+            peer = DistributedPeer(event.connection.username, event.connection)  # type: ignore
             self.distributed_peers.append(peer)
 
             # Only check if the peer is a potential child if the connection
@@ -464,29 +570,41 @@ class DistributedNetwork(BaseManager):
         self._session = None
 
     async def _on_state_changed(self, event: ConnectionStateChangedEvent):
-        if not isinstance(event.connection, PeerConnection):
-            return
-
-        if event.connection.connection_type != PeerConnectionType.DISTRIBUTED:
-            return
-
-        if event.state == ConnectionState.CLOSED:
-            peer = self.get_distributed_peer(
-                event.connection.username, event.connection)
-            if not peer:
-                logger.warning(
-                    f"distributed connection was not registered with the network : {event.connection!r}")
+        connection = event.connection
+        if isinstance(connection, PeerConnection):
+            if connection.connection_type != PeerConnectionType.DISTRIBUTED:
                 return
 
-            # Check if it was the parent or child that was disconnected
-            if self.parent and peer == self.parent:
+            if event.state == ConnectionState.CLOSED:
+                peer = self.get_distributed_peer(connection)
+                # Check if there is a distributed peer registered for this
+                # connection. A distributed peer is only registered once the
+                # full connection initialization has been complete. When the
+                # connection fails before this completion this situation could
+                # occur
+                if not peer:
+                    logger.warning(
+                        f"connection was not registered with the distributed network : {connection!r}")
+                    return
+
+                # Check if it was the parent or child that was disconnected
+                if self.parent and peer == self.parent:
+                    await self._unset_parent()
+
+                if peer in self.children:
+                    self._remove_child(peer)
+
+                # Remove from the distributed connections
+                self.distributed_peers.remove(peer)
+
+        elif isinstance(connection, ServerConnection):
+
+            # When the client itself is branch root it means that the server is
+            # our "parent" and must be unset here
+            if self.parent and self.parent.connection is None:
                 await self._unset_parent()
 
-            if peer in self.children:
-                self._remove_child(peer)
-
-            # Remove from the distributed connections
-            self.distributed_peers.remove(peer)
+            self._reset_server_values()
 
     async def send_messages_to_children(self, *messages: Union[MessageDataclass, bytes]):
         for child in self.children:
