@@ -1,12 +1,15 @@
 from __future__ import annotations
+from aiofiles.threadpool.binary import AsyncBufferedIOBase, AsyncBufferedReader
 import asyncio
+from async_timeout import timeout as atimeout
 from enum import auto, Enum
-from typing import BinaryIO, Callable, List, Optional, TYPE_CHECKING, Union
+from typing import Callable, List, Optional, TYPE_CHECKING, Union
 import logging
 import socket
 import struct
 
 from ..constants import (
+    DISCONNECT_TIMEOUT,
     PEER_CONNECT_TIMEOUT,
     PEER_READ_TIMEOUT,
     SERVER_CONNECT_TIMEOUT,
@@ -28,9 +31,9 @@ from ..protocol.messages import (
     ServerMessage,
 )
 from ..utils import task_counter
+from .rate_limiter import RateLimiter, UnlimitedRateLimiter
 
 if TYPE_CHECKING:
-    from .rate_limiter import RateLimiter
     from .network import Network
 
 
@@ -91,6 +94,9 @@ class Connection:
         self.network: Network = network
         self.state: ConnectionState = ConnectionState.UNINITIALIZED
 
+    async def disconnect(self, reason: CloseReason = CloseReason.UNKNOWN):
+        raise NotImplementedError()
+
     async def set_state(self, state: ConnectionState, close_reason: CloseReason = CloseReason.UNKNOWN):
         self.state = state
         await self.network.on_state_changed(state, self, close_reason=close_reason)
@@ -190,6 +196,9 @@ class DataConnection(Connection):
 
         The connection needs to be established for this method to work
         """
+        if not self._writer:  # pragma: no cover
+            raise RuntimeError("connection not initialized")
+
         return self._writer.get_extra_info('sockname')[0]
 
     async def connect(self, timeout: float = 30):
@@ -208,8 +217,9 @@ class DataConnection(Connection):
         await self.set_state(ConnectionState.CONNECTING)
 
         try:
-            self._reader, self._writer = await asyncio.wait_for(
-                asyncio.open_connection(self.hostname, self.port), timeout)
+            async with atimeout(timeout):
+                self._reader, self._writer = await asyncio.open_connection(
+                    self.hostname, self.port)
 
         except (Exception, asyncio.TimeoutError) as exc:
             await self.disconnect(CloseReason.CONNECT_FAILED)
@@ -239,10 +249,12 @@ class DataConnection(Connection):
             if self._writer is not None:
                 if not self._writer.is_closing():
                     self._writer.close()
-                await self._writer.wait_closed()
+
+                async with atimeout(DISCONNECT_TIMEOUT):
+                    await self._writer.wait_closed()
 
         except Exception as exc:
-            logger.warning(f"{self.hostname}:{self.port} : exception while disconnecting : {exc}")
+            logger.warning(f"{self.hostname}:{self.port} : exception while disconnecting : {exc!r}")
 
         finally:
             await self.set_state(ConnectionState.CLOSED, close_reason=reason)
@@ -262,16 +274,21 @@ class DataConnection(Connection):
 
     # Read/write methods
 
-    async def _read_until_eof(self):
-        await self._reader.read(-1)
+    async def _read_until_eof(self) -> bytes:
+        return await self._reader.read(-1)  # type: ignore[union-attr]
+
+    async def receive_until_eof(self, raise_exception: bool = True) -> Optional[bytes]:
+        if not self._reader:
+            raise ConnectionReadError("cannot read until EOF, connection is not open")
 
-    async def receive_until_eof(self, raise_exception=True):
         try:
             return await self._read(self._read_until_eof)
         except ConnectionReadError:
             if raise_exception:
                 raise
 
+            return None
+
     async def _message_reader_loop(self):
         """Message reader loop. This will loop until the connection is closed or
         the network is closed
@@ -294,16 +311,20 @@ class DataConnection(Connection):
                     if message:
                         await self._perform_message_callback(message)
 
-    async def _read_message(self):
+    async def _read_message(self) -> bytes:
+        if not self._reader:
+            raise ConnectionReadError("cannot read message, connection is not open")
+
         header_size = HEADER_SIZE_OBFUSCATED if self.obfuscated else HEADER_SIZE_UNOBFUSCATED
-        header = await asyncio.wait_for(
-            self._reader.readexactly(header_size), self.read_timeout)
+
+        async with atimeout(self.read_timeout):
+            header = await self._reader.readexactly(header_size)
 
         message_len_buf = obfuscation.decode(header) if self.obfuscated else header
         _, message_len = uint32.deserialize(0, message_len_buf)
 
-        message = await asyncio.wait_for(
-            self._reader.readexactly(message_len), self.read_timeout)
+        async with atimeout(self.read_timeout):
+            message = await self._reader.readexactly(message_len)  #
 
         return header + message
 
@@ -372,14 +393,16 @@ class DataConnection(Connection):
         """
         try:
             if timeout:
-                return await asyncio.wait_for(reader_func(), timeout)
+                async with atimeout(timeout):
+                    return await reader_func()
             else:
                 return await reader_func()
 
         except asyncio.IncompleteReadError as exc:
             if exc.partial:
                 await self.disconnect(CloseReason.READ_ERROR)
-                raise ConnectionReadError(f"{self.hostname}:{self.port} : incomplete read on connection : {exc.partial!r}") from exc
+                raise ConnectionReadError(
+                    f"{self.hostname}:{self.port} : incomplete read on connection : {exc.partial!r}") from exc
             else:
                 await self.disconnect(CloseReason.EOF)
                 return None
@@ -425,6 +448,9 @@ class DataConnection(Connection):
                 f"{self.hostname}:{self.port} : not sending message, connection is closing / closed : {message!r}")
             return
 
+        if not self._writer:
+            raise ConnectionWriteError("cannot send message, connection is not open")
+
         logger.debug(f"{self.hostname}:{self.port} : send message : {message!r}")
         # Serialize the message
         try:
@@ -435,7 +461,8 @@ class DataConnection(Connection):
         # Perform actual send
         try:
             self._writer.write(data)
-            await asyncio.wait_for(self._writer.drain(), 10)
+            async with atimeout(10):
+                await self._writer.drain()
 
         except asyncio.TimeoutError as exc:
             await self.disconnect(CloseReason.TIMEOUT)
@@ -483,8 +510,8 @@ class PeerConnection(DataConnection):
         self.connection_type: str = connection_type
         self.read_timeout = PEER_READ_TIMEOUT
 
-        self.download_rate_limiter: Optional[RateLimiter] = None
-        self.upload_rate_limiter: Optional[RateLimiter] = None
+        self.download_rate_limiter: RateLimiter = UnlimitedRateLimiter()
+        self.upload_rate_limiter: RateLimiter = UnlimitedRateLimiter()
 
     async def connect(self, timeout: float = PEER_CONNECT_TIMEOUT):
         await super().connect(timeout=timeout)
@@ -518,20 +545,36 @@ class PeerConnection(DataConnection):
         self.connection_state = state
 
     async def _read_transfer_ticket(self) -> bytes:
+        if not self._reader:
+            raise ConnectionReadError("cannot read transfer ticket, connection is not open")
+
         return await self._reader.readexactly(struct.calcsize('I'))
 
     async def _read_transfer_offset(self) -> bytes:
+        if not self._reader:
+            raise ConnectionReadError("cannot read transfer offset, connection is not open")
+
         return await self._reader.readexactly(struct.calcsize('Q'))
 
     async def receive_transfer_ticket(self) -> int:
         """Receive the transfer ticket from the connection"""
         data = await self._read(self._read_transfer_ticket)
+
+        if data is None:  # pragma: no cover
+            raise ConnectionReadError(
+                "couldn't receive transfer ticket, connection closed")
+
         _, ticket = uint32.deserialize(0, data)
         return ticket
 
     async def receive_transfer_offset(self) -> int:
         """Receive the transfer offset from the connection"""
         data = await self._read(self._read_transfer_offset)
+
+        if data is None:  # pragma: no cover
+            raise ConnectionReadError(
+                "couldn't receive transfer offset, connection closed")
+
         _, offset = uint64.deserialize(0, data)
         return offset
 
@@ -547,9 +590,12 @@ class PeerConnection(DataConnection):
             socket
         :return: `bytes` object containing the received data
         """
+        if not self._reader:
+            raise ConnectionReadError("cannot read data, connection is not open")
+
         try:
-            data = await asyncio.wait_for(
-                self._reader.read(n_bytes), timeout)
+            async with atimeout(timeout):
+                data = await self._reader.read(n_bytes)
 
         except asyncio.TimeoutError as exc:
             await self.disconnect(CloseReason.TIMEOUT)
@@ -566,7 +612,8 @@ class PeerConnection(DataConnection):
             else:
                 return data
 
-    async def receive_file(self, file_handle: BinaryIO, filesize: int, callback: Optional[Callable[[bytes], None]] = None):
+    async def receive_file(
+            self, file_handle: AsyncBufferedIOBase, filesize: int, callback: Optional[Callable[[bytes], None]] = None):
         """Receives a file on the current connection and writes it to the given
         `file_handle`
 
@@ -595,9 +642,13 @@ class PeerConnection(DataConnection):
                 return
 
     async def send_data(self, data: bytes):
+        if not self._writer:
+            raise ConnectionWriteError("cannot send data, connection is not open")
+
         try:
             self._writer.write(data)
-            await asyncio.wait_for(self._writer.drain(), TRANSFER_TIMEOUT)
+            async with atimeout(TRANSFER_TIMEOUT):
+                await self._writer.drain()
 
         except asyncio.TimeoutError as exc:
             await self.disconnect(CloseReason.TIMEOUT)
@@ -607,7 +658,7 @@ class PeerConnection(DataConnection):
             await self.disconnect(CloseReason.WRITE_ERROR)
             raise ConnectionWriteError(f"{self.hostname}:{self.port} : write error") from exc
 
-    async def send_file(self, file_handle: BinaryIO, callback: Optional[Callable[[bytes], None]] = None):
+    async def send_file(self, file_handle: AsyncBufferedReader, callback: Optional[Callable[[bytes], None]] = None):
         """Sends a file over the connection. This method makes use of the
         `upload_rate_limiter` to limit how many bytes are being sent at a time.
 
