@@ -1,5 +1,6 @@
 from aiofiles import os as asyncos
 import asyncio
+from concurrent.futures import Executor
 from functools import partial
 import logging
 import mutagen
@@ -8,13 +9,26 @@ import os
 import re
 import sys
 import time
-from typing import Optional, Dict, List, Set, Tuple, Union
+from typing import (
+    Callable,
+    Dict,
+    List,
+    Optional,
+    Set,
+    Tuple,
+    Union,
+)
 import uuid
 from weakref import WeakSet
 
 from ..base_manager import BaseManager
 from .cache import SharesNullCache, SharesCache
-from ..events import EventBus, ScanCompleteEvent, SessionInitializedEvent
+from ..events import (
+    EventBus,
+    ScanCompleteEvent,
+    SessionInitializedEvent,
+    SessionDestroyedEvent,
+)
 from ..exceptions import (
     FileNotFoundError,
     FileNotSharedError,
@@ -30,11 +44,14 @@ from ..network.network import Network
 from ..protocol.messages import SharedFoldersFiles
 from ..protocol.primitives import DirectoryData
 from ..search.model import SearchQuery
+from ..session import Session
 from ..settings import Settings
-from .utils import create_term_pattern, convert_items_to_file_data
+from .utils import convert_items_to_file_data
 
 logger = logging.getLogger(__name__)
 
+ItemAttributes = List[Tuple[int, int]]
+ExecutorFactory = Callable[[], Executor]
 
 _COMPRESSED_FORMATS = [
     'MP3',
@@ -50,9 +67,17 @@ _QUERY_CLEAN_PATTERN = re.compile(r"[\W_]")
 """Pattern to remove all non-word/digit characters from a string"""
 
 
-def scan_directory(shared_directory: SharedDirectory, children: Optional[List[SharedDirectory]] = None) -> Set[SharedItem]:
+def scan_directory(
+        shared_directory: SharedDirectory,
+        children: Optional[List[SharedDirectory]] = None) -> Set[SharedItem]:
     """Scans the directory for items to share
 
+    Warning: when using ProcessPoolExecutor on this method the returned items
+    will not have the shared directory set in some cases. The shared_directory
+    passed here is a copy and the object will be removed once this function
+    finishes (see https://stackoverflow.com/a/72726998/1419478). In this case
+    you should manually assign it again
+
     :param shared_directory: `SharedDirectory` instance
     :param children: list of `SharedDirectory` instances, the items in this list
         will not be returned and should be scanned individually
@@ -86,7 +111,7 @@ def scan_directory(shared_directory: SharedDirectory, children: Optional[List[Sh
     return shared_items
 
 
-def extract_attributes(filepath: str) -> List[Tuple[int, int]]:
+def extract_attributes(filepath: str) -> ItemAttributes:
     """Attempts to extract attributes from the file at `filepath`. If there was
     an error attempting to extract the attributes this method will log a warning
     and return an empty list of attributes
@@ -129,25 +154,31 @@ class SharesManager(BaseManager):
 
     def __init__(
             self, settings: Settings, event_bus: EventBus,
-            network: Network, cache: Optional[SharesCache] = None):
+            network: Network, cache: Optional[SharesCache] = None,
+            executor_factory: Optional[ExecutorFactory] = None):
         self._settings: Settings = settings
         self._event_bus: EventBus = event_bus
         self._network: Network = network
         self._term_map: Dict[str, WeakSet[SharedItem]] = {}
-        self._shared_directories: List[SharedDirectory] = list()
-        self.scan_task: Optional[asyncio.Task] = None
+        self._shared_directories: List[SharedDirectory] = []
+        self._session: Optional[Session] = None
 
         self.cache: SharesCache = cache if cache else SharesNullCache()
-        self.executor = None
+        self.executor: Optional[Executor] = None
+        self.executor_factory: Optional[ExecutorFactory] = executor_factory
 
         self.naming_strategies = [
             DefaultNamingStrategy(),
             NumberDuplicateStrategy()
         ]
 
+        self.register_listeners()
+
     def register_listeners(self):
         self._event_bus.register(
             SessionInitializedEvent, self._on_session_initialized)
+        self._event_bus.register(
+            SessionDestroyedEvent, self._on_session_destroyed)
 
     @property
     def shared_directories(self):
@@ -187,6 +218,17 @@ class SharesManager(BaseManager):
 
         return alias_string
 
+    async def start(self):
+        if self.executor_factory:
+            self.executor = self.executor_factory()
+
+    async def stop(self) -> List[asyncio.Task]:
+        if self.executor:
+            self.executor.shutdown()
+            self.executor = None
+
+        return []
+
     async def load_data(self):
         self.read_cache()
         self.load_from_settings()
@@ -195,13 +237,34 @@ class SharesManager(BaseManager):
         self.write_cache()
 
     def load_from_settings(self):
-        """Loads the directories from the settings"""
-        for shared_directory in self._settings.shares.directories:
-            self.add_shared_directory(
-                shared_directory.path,
-                share_mode=shared_directory.share_mode,
-                users=shared_directory.users
-            )
+        """Loads the shared directories from the settings. Existing directories
+        will be updated, non-existing directories added, directories that no
+        longer exist will be removed
+        """
+        new_shared_directories: List[SharedDirectory] = []
+
+        # Add or update directories
+        for directory_entry in self._settings.shares.directories:
+            try:
+                shared_directory = self.get_shared_directory(directory_entry.path)
+            except SharedDirectoryError:
+                shared_directory = self.add_shared_directory(
+                    directory_entry.path,
+                    share_mode=directory_entry.share_mode,
+                    users=directory_entry.users
+                )
+            else:
+                self.update_shared_directory(
+                    shared_directory,
+                    share_mode=directory_entry.share_mode,
+                    users=directory_entry.users or []
+                )
+
+            new_shared_directories.append(shared_directory)
+
+        self._shared_directories = new_shared_directories
+
+        self.rebuild_term_map()
 
     def read_cache(self):
         """Read the directories from the cache"""
@@ -246,14 +309,15 @@ class SharesManager(BaseManager):
     async def get_shared_item(self, remote_path: str, username: Optional[str] = None) -> SharedItem:
         """Gets a shared item from the cache based on the given file path. If
         the file does not exist in the `shared_items` or the file is present
-        in the cache but does not exist on disk a `FileNotFoundError` is raised
+        in the cache but does not exist on disk a :class:`.FileNotFoundError` is
+        raised
 
         If a `username` is passed this will also check if the file is locked
-        and raise a `FileNotSharedError` if the file is not accessible for that
-        user
+        and raise a :class:`.FileNotSharedError` if the file is not accessible
+        for that user
 
         :param remote_path: the remote_path
-        :param username:
+        :param username: optional username to check if the file is shared or not
         :raise FileNotFoundError: filename was not found in shared_items or was found
             but did not exist on disk
         :raise FileNotSharedError: file is found, but locked for the given
@@ -278,21 +342,28 @@ class SharesManager(BaseManager):
 
     def add_shared_directory(
             self, shared_directory: str,
-            share_mode: DirectoryShareMode = DirectoryShareMode.EVERYONE, users: Optional[List[str]] = None) -> SharedDirectory:
+            share_mode: DirectoryShareMode = DirectoryShareMode.EVERYONE,
+            users: Optional[List[str]] = None) -> SharedDirectory:
         """Adds a shared directory. This method will call `generate_alias` and
-        add the directory to the directory map.
+        add the directory to the directory map. This method will not scan the
+        directory, for scanning see the :meth:`scan`, :meth:`scan_directory_files`
+        and :meth:`scan_directory_file_attributes` methods.
 
         :param shared_directory: path of the shared directory
         :param share_mode: the share mode for the directory
         :param users: in case the share mode is `USERS`, a list of users to
             share it with
-        :return: a `SharedDirectory` object
+        :return: a :class:`.SharedDirectory` object
+        :raise SharedDirectoryError: if the directory is already shared
         """
-        # Calc absolute path, generate an alias and store it
+        if self.is_directory_shared(shared_directory):
+            raise SharedDirectoryError(
+                f"directory {shared_directory} is already shared")
+
+        # Generate an absolute path, alias and create the object
         abs_directory = os.path.normpath(os.path.abspath(shared_directory))
         alias = self.generate_alias(abs_directory)
 
-        # Check if we have an existing shared directory, otherwise return it
         directory_object = SharedDirectory(
             shared_directory,
             abs_directory,
@@ -300,9 +371,6 @@ class SharesManager(BaseManager):
             share_mode=share_mode,
             users=users or []
         )
-        for shared_dir in self._shared_directories:
-            if shared_dir == directory_object:
-                return shared_dir
 
         # If the new directory is a child of an existing directory, move items
         # to the child directory and remove them from the parent directory
@@ -316,35 +384,114 @@ class SharesManager(BaseManager):
         self._shared_directories.append(directory_object)
         return directory_object
 
-    def remove_shared_directory(self, directory: SharedDirectory):
+    def update_shared_directory(
+            self, directory: Union[str, SharedDirectory],
+            share_mode: Optional[DirectoryShareMode] = None,
+            users: Optional[List[str]] = None) -> SharedDirectory:
+        """Updates `share_mode` and `users` values for the given directory
+
+        :param directory: if a string is given this method will attempt to find
+            the shared directory based on the absolute path
+        :return: the updated directory
+        :raise SharedDirectoryError: if the given directory is a string and not
+            in the list of shared directories
+        """
+        if isinstance(directory, str):
+            shared_directory = self.get_shared_directory(directory)
+        else:
+            shared_directory = directory
+
+        if share_mode is not None:
+            shared_directory.share_mode = share_mode
+
+        if users is not None:
+            shared_directory.users = users
+
+        return shared_directory
+
+    def remove_shared_directory(self, directory: Union[str, SharedDirectory]) -> SharedDirectory:
         """Removes the given shared directory. If the directory was a
         subdirectory of another shared directory its items will be moved into
         that directory
 
-        :param shared_directory: `SharedDirectory` instance to remove
-        :raise SharedDirectoryError: raised when the passed `shared_directory`
-            was not added to the manager
+        Example file structure:
+
+        * Music\\
+
+            * Artist_One\\
+
+                * song_one.mp3
+
+            * Artist_Two\\
+
+                * song_two.mp3
+
+        If the user shares 2 directories:
+
+        * Music\\ : EVERYONE
+        * Music\\Artist_One\\ : FRIENDS
+
+        And removes the `Music\\Artist_One\\ shared directory the files of that
+        directory get returned back to the parent directory. In this case file
+        `song_two.mp3` will be shared with EVERYONE again
+
+        :param shared_directory: :class:`.SharedDirectory` instance to remove
+        :return: the removed directory
+        :raise SharedDirectoryError: raised when the passed `directory` was not
+            added to the manager
         """
-        if directory not in self._shared_directories:
-            raise SharedDirectoryError(
-                "attempted to remove directory which was not added to the manager"
-            )
+        if isinstance(directory, str):
+            shared_directory = self.get_shared_directory(directory)
+        else:
+            if directory not in self._shared_directories:
+                raise SharedDirectoryError(
+                    "attempted to remove directory which was not added to the manager"
+                )
+            else:
+                shared_directory = directory
 
-        self._shared_directories.remove(directory)
+        self._shared_directories.remove(shared_directory)
 
-        parents = self._get_parent_directories(directory)
+        parents = self._get_parent_directories(shared_directory)
         # If the directory has a parent directory, move all items into that
         # directory
         if parents:
             parent = parents[-1]
-            parent.items |= directory.items
+            parent.items |= shared_directory.items
 
         self._cleanup_term_map()
 
+        return shared_directory
+
+    def get_shared_directory(self, directory: str) -> SharedDirectory:
+        """Calculates the absolute path of given `directory` and looks for the
+        matching :class:`SharedDirectory` instance
+
+        :raise SharedDirectoryError: if the directory is not found
+        """
+        abs_path = os.path.normpath(os.path.abspath(directory))
+        for shared_directory in self.shared_directories:
+            if shared_directory.absolute_path == abs_path:
+                return shared_directory
+
+        raise SharedDirectoryError(
+            f"did not find shared directory with path : {directory}")
+
+    def is_directory_shared(self, directory: str) -> bool:
+        """Checks if a directory is already a shared directory by checking the
+        absolute path of that directory
+        """
+        try:
+            self.get_shared_directory(directory)
+        except SharedDirectoryError:
+            return False
+        else:
+            return True
+
     async def scan_directory_files(self, shared_directory: SharedDirectory):
         """Scans the files for the given `shared_directory`
 
-        :param shared_directory: `SharedDirectory` instance to scan
+        :param shared_directory: :class:`.SharedDirectory` instance to scan
         :raise SharedDirectoryError: raised when the passed `shared_directory`
             was not added to the manager
         """
@@ -370,10 +517,20 @@ class SharesManager(BaseManager):
         else:
             logger.debug(f"scan found {len(shared_items)} files for directory {shared_directory!r}")
 
+            # When using a ProcessPoolExecutor the `shared_directory` property
+            # might be set to None (since objects are cloned instead of passed
+            # by reference). Re-assign it the proper object
+            for item in shared_items:
+                item.shared_directory = shared_directory
+
             # Adds all new items to the directory items
             shared_directory.items |= shared_items
+
             # Remove all items from the cache that weren't in the returned items
-            # set
+            # set. Because the `modified` parameter is part of the hash items
+            # with a changed `modified` parameter will also be removed meaning
+            # their attributes will be reset and these files attributes need
+            # to be rescanned
             shared_directory.items -= (shared_directory.items ^ shared_items)
 
         self._build_term_map(shared_directory)
@@ -384,45 +541,52 @@ class SharesManager(BaseManager):
         only files that do not yet have attributes will be scanned
 
         The results of the scan are handled internally and are automatically to
-        the `SharedItem` object for which the scan was performed
+        the :class:`.SharedItem` object for which the scan was performed
 
-        :param shared_directory: `SharedDirectory` instance for which the files
-            need to be scanned
+        :param shared_directory: :class:`.SharedDirectory` instance for which
+            the files need to be scanned
         :return: List of futures for each file that needs to be scanned
         """
-        def extract_item_attributes(item: SharedItem):
-            return item, extract_attributes(item.get_absolute_path())
-
         loop = asyncio.get_running_loop()
 
         # Schedule the items on the executor
-        extract_futures: List[asyncio.Future] = []
+        futures: List[asyncio.Future] = []
         for item in shared_directory.items:
             if item.attributes is None:
                 future = loop.run_in_executor(
                     self.executor,
-                    partial(extract_item_attributes, item)
+                    partial(extract_attributes, item.get_absolute_path())
                 )
-                extract_futures.append(future)
+                future.add_done_callback(
+                    partial(self._extract_attributes_callback, item)
+                )
+                futures.append(future)
 
         logger.debug(
-            f"scheduled {len(extract_futures)} / {len(shared_directory.items)} items "
+            f"scheduled {len(futures)} / {len(shared_directory.items)} items "
             f"for attribute extracting for directory {shared_directory}")
 
-        for future in asyncio.as_completed(extract_futures):
-            try:
-                item, attributes = await future
-            except Exception:
-                logger.warning(f"exception fetching shared item attributes")
-            else:
-                item.attributes = attributes
+        start = time.perf_counter()
+
+        await asyncio.gather(*futures, return_exceptions=True)
+
+        logger.debug(
+            f"scanned attributes for {len(futures)} items in {time.perf_counter() - start}s")
+
+    def _extract_attributes_callback(self, item: SharedItem, future: asyncio.Future):
+        try:
+            attributes = future.result()
+        except Exception:
+            logger.warning("exception fetching shared item attributes")
+        else:
+            item.attributes = attributes
 
     async def scan(self):
         """Scan the files and their attributes for all directories currently
         defined in the `shared_directories`
 
-        This method will emit a `ScanCompleteEvent` on the event bus and report
-        the shares to the server
+        This method will emit a :class:`.ScanCompleteEvent` on the event bus
+        and report the shares to the server
         """
         start_time = time.perf_counter()
 
@@ -436,7 +600,7 @@ class SharesManager(BaseManager):
             self.scan_directory_file_attributes(shared_directory)
             for shared_directory in self._shared_directories
         ]
-        await asyncio.gather(*attribute_futures, return_exceptions=True)
+        await asyncio.gather(*attribute_futures)
 
         logger.info(f"completed scan in {time.perf_counter() - start_time} seconds")
         folder_count, file_count = self.get_stats()
@@ -449,7 +613,10 @@ class SharesManager(BaseManager):
     async def get_filesize(self, shared_item: SharedItem) -> int:
         return await asyncos.path.getsize(shared_item.get_absolute_path())
 
-    def query(self, query: Union[str, SearchQuery], username: Optional[str] = None) -> Tuple[List[SharedItem], List[SharedItem]]:
+    def query(
+            self, query: Union[str, SearchQuery],
+            username: Optional[str] = None,
+            excluded_search_phrases: Optional[List[str]] = None) -> Tuple[List[SharedItem], List[SharedItem]]:
         """Performs a query on the `shared_directories` returning the matching
         items. If `username` is passed this method will return a list of
         visible results and list of locked results. If `None` the second list
@@ -462,6 +629,8 @@ class SharesManager(BaseManager):
         :param username: optionally the username of the user making the query.
             This is used to determine locked results, if not given the locked
             results list will be empty
+        :param excluded_search_phrases: optional list of search phrases that
+            should be excluded from the search results
         :return: two lists of visible results and locked results
         """
         if not isinstance(query, SearchQuery):
@@ -512,31 +681,31 @@ class SharesManager(BaseManager):
         for include_term in include_terms:
             found_items &= set(self._term_map[include_term])
 
-        # Regular expressions on the remnants
+        # Regular expressions on the remaining items
 
-        for include_term in search_query.include_terms:
-            to_remove = set()
-            for item in found_items:
-                pattern = create_term_pattern(include_term, wildcard=False)
-                if not re.search(pattern, item.get_query_path()):
-                    to_remove.add(item)
-            found_items -= to_remove
+        to_keep = set()
+        excl_phrases = excluded_search_phrases or []
+        for found_item in found_items:
 
-        for wildcard_term in search_query.wildcard_terms:
-            to_remove = set()
-            for item in found_items:
-                pattern = create_term_pattern(wildcard_term, wildcard=True)
-                if not re.search(pattern, item.get_query_path()):
-                    to_remove.add(item)
-            found_items -= to_remove
+            if not all(matcher(found_item.get_query_path()) for matcher in search_query.matchers_iter()):
+                continue
 
-        for exclude_term in search_query.exclude_terms:
-            to_remove = set()
-            for item in found_items:
-                pattern = create_term_pattern(exclude_term, wildcard=False)
-                if re.search(pattern, item.get_query_path()):
-                    to_remove.add(item)
-            found_items -= to_remove
+            # Excluded search phrases
+            for excl_phrase in excl_phrases:
+                if excl_phrase in found_item.get_query_path().lower():
+                    logger.debug(
+                        f"removing search result '{found_item.get_absolute_path()}' due to "
+                        f"excluded phrase '{excl_phrase}'"
+                    )
+                    break
+
+            else:
+                to_keep.add(found_item)
+
+            if len(to_keep) >= self._settings.searches.max_results:
+                break
+
+        found_items = to_keep
 
         # Order by visible and locked results when username is given
         if username:
@@ -592,30 +761,35 @@ class SharesManager(BaseManager):
 
     def create_shares_reply(self, username: str) -> Tuple[List[DirectoryData], List[DirectoryData]]:
         """Creates a complete list of the currently shared items as a reply to
-        a `PeerSharesRequest` messages
+        a :class:`.PeerSharesRequest` messages
 
         :param username: username of the user requesting the shares reply, this
             is used to determine the locked results
         :return: tuple with two lists: public directories and locked directories
         """
-        def list_unique_directories(directories: List[SharedDirectory]) -> Dict[str, List[SharedItem]]:
-            # Sort files under unique directories by path
-            response_dirs: Dict[str, List[SharedItem]] = {}
+        def list_unique_directories(directories: List[SharedDirectory]) -> Dict[Tuple[str, ...], List[SharedItem]]:
+            response_dirs: Dict[Tuple[str, ...], List[SharedItem]] = {}
+
             for directory in directories:
                 for item in directory.items:
-                    dir_path = item.get_remote_directory_path()
-                    if dir_path in response_dirs:
-                        response_dirs[dir_path].append(item)
-                    else:
-                        response_dirs[dir_path] = [item, ]
+                    dir_path = item.get_remote_directory_path_parts()
+
+                    # Create all possible subdirectory paths
+                    for x in range(len(dir_path)):
+                        dir_path_subpart = dir_path[:x + 1]
+                        if dir_path_subpart not in response_dirs:
+                            response_dirs[dir_path_subpart] = []
+
+                    response_dirs[dir_path].append(item)
+
             return response_dirs
 
-        def convert_to_directory_shares(directory_map: Dict[str, List[SharedItem]]) -> List[DirectoryData]:
+        def convert_to_directory_shares(directory_map: Dict[Tuple[str, ...], List[SharedItem]]) -> List[DirectoryData]:
             public_shares = []
             for directory, files in directory_map.items():
                 public_shares.append(
                     DirectoryData(
-                        name=directory,
+                        name='\\'.join(directory),
                         files=convert_items_to_file_data(files, use_full_path=False)
                     )
                 )
@@ -630,34 +804,43 @@ class SharesManager(BaseManager):
         return visible_shares, locked_shares
 
     def create_directory_reply(self, remote_directory: str) -> List[DirectoryData]:
-        """Lists directory data as a response to a directory request.
+        """Lists directory data as a response to a directory request. This will
+        not contain any information about subdirectories, only the files within
+        that directory
+
+        This method differs from the official clients who don't respond at all
+        to the message if the directory does not exist. Instead if the directory
+        does not exist and empty list is returned
 
         :param remote_directory: remote path of the directory
+        :return: list of directories. Empty if the directory is not shared, a
+            list with one entry if the directory is found
         """
-        remote_directory = remote_directory.rstrip('\\/')
-        response_dirs: Dict[str, List[SharedItem]] = {}
+        items: List[SharedItem] = []
+
+        remote_dir_parts = tuple(remote_directory.split('\\'))
+        remote_dir_parts_len = len(remote_dir_parts)
+        is_shared = False
+
         for shared_dir in self._shared_directories:
             for item in shared_dir.items:
-                item_dir = item.get_remote_directory_path()
-                if item_dir != remote_directory:
-                    continue
+                # Funny logic to determine if the directory is shared
+                if not is_shared:
+                    if item.get_remote_directory_path_parts()[:remote_dir_parts_len] == remote_dir_parts:
+                        is_shared = True
 
-                dir_name = shared_dir.get_remote_path()
-                if dir_name in response_dirs:
-                    response_dirs[dir_name].append(item)
-                else:
-                    response_dirs[dir_name] = [item, ]
+                if item.get_remote_directory_path() == remote_directory:
+                    items.append(item)
 
-        reply = []
-        for directory, files in response_dirs.items():
-            reply.append(
-                DirectoryData(
-                    name=directory,
-                    files=convert_items_to_file_data(files, use_full_path=False)
-                )
+        if not is_shared:
+            return []
+
+        return [
+            DirectoryData(
+                name=remote_directory,
+                files=convert_items_to_file_data(items, use_full_path=False)
             )
-
-        return reply
+        ]
 
     def _add_item_to_term_map(self, item: SharedItem):
         path = (item.subdir + "/" + item.filename).lower()
@@ -708,6 +891,10 @@ class SharesManager(BaseManager):
 
     async def report_shares(self):
         """Reports the shares amount to the server"""
+        if not self._session:
+            logger.warning("attempted to report shares without valid session")
+            return
+
         folder_count, file_count = self.get_stats()
         logger.debug(f"reporting shares ({folder_count=}, {file_count=})")
         await self._network.send_server_messages(
@@ -718,4 +905,8 @@ class SharesManager(BaseManager):
         )
 
     async def _on_session_initialized(self, event: SessionInitializedEvent):
+        self._session = event.session
         await self.report_shares()
+
+    async def _on_session_destroyed(self, event: SessionDestroyedEvent):
+        self._session = None
